// ExtentCalcs.cpp: implementation of the CExtentCalcs class.
// Copyright (c) 2015 by David Wincelberg
//
// Reference:
// "How to implement horizontal scrolling in a CListBox-derived class" at:
// https://support.microsoft.com/en-us/kb/146437
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ExtentCalcs.h"

#include "ListBoxData.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CExtentCalcs::CExtentCalcs()
{
	m_nMaxExtent  = 0;
	m_bDefer      = false;		// Used in DeleteString() and ResetContent().
	m_bNeedRecalc = false;		// Used in DeleteString() and ResetContent().
	m_nOffset     = 0;			// Amount to add to extents.  Used by CSortCheckListBox.
	m_nTabStops   = 0;			// no tab stops
	m_pTabStops   = nullptr;	// array of tab stops
}


CExtentCalcs::~CExtentCalcs()
{
	delete m_pTabStops;
}


// Action:	Sets a pointer to the list box and allows for an extent offset.
void CExtentCalcs::SetBoxPointer (CListBox* pBox, int nOffset /* = 0 */)
{
	ASSERT (pBox);
	m_pBox    = pBox;
	m_nOffset = nOffset;
}	// SetBoxPointer


// Action:	Stores the horizontal extent in the stored data.
void CExtentCalcs::StoreExtent (int nItem, LPCTSTR lpszItem, CListBoxData* pListBoxData)
{
	if (lpszItem == nullptr || pListBoxData == nullptr)
		return;

	ASSERT (m_pBox);
	CDC* pDC = m_pBox->GetDC();
	if (pDC == nullptr)
		return;

	CFont* pFont = m_pBox->GetFont();
	ASSERT (pFont);
	CFont* pPrevFont = pDC->SelectObject (pFont);

	int nExtent = CalcHorizontalExtent (lpszItem, pDC);

	pDC->SelectObject (pPrevFont);
	m_pBox->ReleaseDC (pDC);

	pListBoxData->SetExtent (nExtent);

	if (nExtent > m_nMaxExtent) {
		m_nMaxExtent = nExtent;
		m_pBox->SetHorizontalExtent (m_nMaxExtent);
	}

	if (nItem >= 0)		// Needed by CSortListBoxOD.
		m_pBox->SetItemData (nItem, reinterpret_cast<DWORD> (pListBoxData));
}	// StoreExtent


// Action:	Calculates the horizontal size of the input string.
// Note:	Based on the above reference.
int CExtentCalcs::CalcHorizontalExtent (LPCTSTR lpszItem, CDC* pDC)
{
	int nExtent = 0;
	CSize size;

	if (m_pBox->GetStyle() & LBS_USETABSTOPS) {
		if (m_pTabStops == nullptr) {
			int nDefault = 2;
			m_pBox->SetTabStops (1, &nDefault);	// Works better than no-param version.
		}

		size = pDC->GetTabbedTextExtent (lpszItem, ::lstrlen (lpszItem),
			m_nTabStops, m_pTabStops);
	}
	else
		size = pDC->GetTextExtent (lpszItem, ::lstrlen (lpszItem));

	pDC->LPtoDP (&size);				// logical units to pixels
	nExtent = size.cx + 6 + m_nOffset;	// adjustment in sample code plus offset

	if (nExtent < 0 || nExtent > 32767)		// useful limits
		nExtent = 32767;	// == 2^15 - 1

	return nExtent;
}	// CalcHorizontalExtent


// Action:	Defers or may call updating the maximum extent.
void CExtentCalcs::DeferExtentCalc (bool bDefer)
{
	if (!bDefer && m_bNeedRecalc)
		UpdateMaxExtent();

	m_bDefer      = bDefer;
	m_bNeedRecalc = false;	// Clears for bracketing code with this function.
}	// DeferExtentCalc


// Action:	Updates the maximum extent based on the stored extents.
void CExtentCalcs::UpdateMaxExtent()
{
	ASSERT (m_pBox);
	int nCount = m_pBox->GetCount();
	if (nCount == LB_ERR)
		return;

	CListBoxData* pListBoxData;
	m_nMaxExtent = 0;

	for (int n = 0;  n < nCount;  ++n) {
		pListBoxData = reinterpret_cast<CListBoxData*> (m_pBox->GetItemData (n));
		if (pListBoxData != nullptr) {
			m_nMaxExtent = max (m_nMaxExtent, pListBoxData->GetExtent());
		}
	}

	m_pBox->SetHorizontalExtent (m_nMaxExtent);
}	// UpdateMaxExtent


// Action:	Recalculates and sets the listbox's horizontal extent.
// Note:	Based on the above reference.
void CExtentCalcs::RecalcHorizontalExtent()
{
	ASSERT (m_pBox);
	int nCount = m_pBox->GetCount();
	if (nCount == LB_ERR)
		return;

	CDC* pDC = m_pBox->GetDC();
	if (pDC == nullptr)
		return;

	CFont* pFont = m_pBox->GetFont();
	ASSERT (pFont);
	CFont* pPrevFont = pDC->SelectObject (pFont);

	CString strItem;
	int nExtent;
	CListBoxData* pListBoxData;
	m_nMaxExtent = 0;

	try {
		for (int n = 0;  n < nCount;  ++n) {
			m_pBox->GetText (n, strItem);

			nExtent = CalcHorizontalExtent (strItem, pDC);

			m_nMaxExtent = max (m_nMaxExtent, nExtent);
			pListBoxData  = reinterpret_cast<CListBoxData*> (m_pBox->GetItemData (n));
			if (pListBoxData != nullptr) {
				pListBoxData->SetExtent (nExtent);
			}
		}
	}
	catch (CMemoryException* e) {	// GetText
		e->Delete();
	}

	pDC->SelectObject (pPrevFont);
	m_pBox->ReleaseDC (pDC);
	m_pBox->SetHorizontalExtent (m_nMaxExtent);
}	// RecalcHorizontalExtent


// Action:	Recalculates extents if the font resources have changed.
// Note:	The top-level window receives the OnFontChange message.
//			It should pass it to listbox controls with extent calculations.
void CExtentCalcs::FontsChanged()
{
	RecalcHorizontalExtent();
}	// FontsChanged


// Action:	Handles the LB_SETTABSTOPS message to update the tab stop array.
// Note:	Based on the above reference.
LRESULT CExtentCalcs::TabStopsSet (WPARAM wParam, LPARAM lParam, LRESULT lResult)
{
	ASSERT (m_pBox);
	m_nTabStops = wParam;

	if (m_pTabStops != nullptr) {
		delete[] m_pTabStops;
		m_pTabStops = nullptr;
	}

	if (m_nTabStops > 0) {
		m_pTabStops = new int[m_nTabStops];
		::memcpy (m_pTabStops, reinterpret_cast<void*> (lParam), m_nTabStops * sizeof (int));

		CDC* pDC = m_pBox->GetDC();
		if (pDC == nullptr)
			return lResult;

		CFont* pFont = m_pBox->GetFont();
		ASSERT (pFont);
		CFont* pPrevFont = pDC->SelectObject (pFont);
		CSize size;

		GetTextExtentPoint32 (pDC->GetSafeHdc(),
			_T("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"),
			52, &size);

		pDC->SelectObject (pPrevFont);
		m_pBox->ReleaseDC (pDC);

		int aveCharWidth = (size.cx / 26 + 1) / 2;

		for (int i = 0;  i < m_nTabStops;  ++i)
			m_pTabStops[i] = (m_pTabStops[i] * aveCharWidth + 2) / 4;
	}

	if (!m_bDefer)
		RecalcHorizontalExtent();

	return lResult;
}	// TabStopsSet
