// SortComboBox.cpp : implementation file
// Copyright (c) 2016 by David Wincelberg
//
// Window styles:
// Recommended: vertical scroll
// Optional: sort
//
// For an owner-draw combo box with strings, select has-strings and
// implement DrawItem and MeasureItem.  CompareItem will not be called.
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SortComboBox.h"

#include "ComboBoxData.h"
#include "ComboBoxOrder.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSortComboBox

// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortComboBox::CSortComboBox (CComboBoxOrder* pComboBoxOrder /* = nullptr */,
							  bool bHandleDelete /* = true */)
{
	m_pComboBoxOrder = (pComboBoxOrder == nullptr ? new CComboBoxOrder() : pComboBoxOrder);
	m_bHandleDelete  = bHandleDelete;
	m_bSorting       = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;
}	// CSortComboBox


// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortComboBox::CSortComboBox (CStringOrder* pStringOrder,
							  bool bHandleDeleteInner /* = true */,
							  bool bHandleDeleteOuter /* = true */)
{
	m_pComboBoxOrder = new CComboBoxOrder (pStringOrder, bHandleDeleteInner);
	m_bHandleDelete  = bHandleDeleteOuter;
	m_bSorting       = false;
	m_bSorted        = true;
	m_bAllowOrderMod = true;
}	// CSortListBox


CSortComboBox::~CSortComboBox()
{
	if (m_bHandleDelete)
		delete m_pComboBoxOrder;
}	// ~CSortComboBox


BEGIN_MESSAGE_MAP(CSortComboBox, CComboBox)
	//{{AFX_MSG_MAP(CSortComboBox)
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSortComboBox message handlers

// Action:	Checks the styles and sets the sorting flag.
void CSortComboBox::PreSubclassWindow()
{
	DWORD dwStyle = GetStyle();
	if ((dwStyle & (CBS_OWNERDRAWFIXED | CBS_OWNERDRAWVARIABLE)) != 0
	&& (dwStyle & CBS_HASSTRINGS) == 0) {
		TRACE (_T("** For CSortComboBox, set style has-strings with an owner-draw style.\n"));
		ASSERT (false);
	}

	m_bSorting = ((dwStyle & CBS_SORT) != 0);

	CComboBox::PreSubclassWindow();
}	// PreSubclassWindow


// Action:	Replaces the sort-order object.
//			Deletes the old one unless m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortComboBox::SetSortOrder (CComboBoxOrder* pComboBoxOrder,
								  bool bHandleDelete /* = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the sort-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pComboBoxOrder == nullptr)
		return;

	if (m_bHandleDelete)
		delete m_pComboBoxOrder;

	m_pComboBoxOrder = pComboBoxOrder;
	m_bHandleDelete  = bHandleDelete;
}	// SetSortOrder


// Action:	Replaces the string-order object.
//			Deletes the old one unless CComboBoxOrder's m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortComboBox::SetStringOrder (CStringOrder* pStringOrder,
									bool bHandleDelete /*  = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the string-order strategy is not allowed ")
			_T("after adding or inserting a string.\n"));
		ASSERT (false);
		return;
	}

	if (pStringOrder == nullptr || m_pComboBoxOrder == nullptr)
		return;

	m_pComboBoxOrder->SetStringOrder (pStringOrder, bHandleDelete);
}	// SetStringOrder


// Action:	Checks if a combobox-data object was supplied; o.w. allocates one.
bool CSortComboBox::CheckComboBoxData (CComboBoxData** ppComboBoxData)
{
	if (*ppComboBoxData == nullptr) {
		try {
			*ppComboBoxData = new CComboBoxData();
		}
		catch (CMemoryException* e) {
			e->Delete();
			return false;
		}
	}

	return true;
}	// CheckComboBoxData


// Action:	Adds a string to the combo box.
//			If sorting is selected, sorts based on the current string order.
// Note:	Allows for storing a data item that is in a subclass of CComboBoxData.
int CSortComboBox::AddString (LPCTSTR lpszItem, CComboBoxData* pComboBoxData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pComboBoxData;
		return CB_ERR;
	}

	if (!CheckComboBoxData (&pComboBoxData))
		return CB_ERR;

	if (!m_bSorting) {
		int nResult = CComboBox::AddString (lpszItem);
		if (nResult == CB_ERR || nResult == CB_ERRSPACE) {
			delete pComboBoxData;
			return nResult;
		}

		CComboBox::SetItemData (nResult, reinterpret_cast<DWORD> (pComboBoxData));
		m_bAllowOrderMod = false;
		return nResult;
	}

	int nPosition = 0, nCount = GetCount();

	if (nCount > 0 && m_pComboBoxOrder != nullptr) {
		try {
			CString strItem = lpszItem;
			nPosition = SortStrings (strItem, pComboBoxData, nCount);
		}
		catch (CMemoryException* e) {
			e->Delete();
			return CB_ERR;
		}
	}

	int nResult = CComboBox::InsertString (nPosition, lpszItem);
	if (nResult == CB_ERR || nResult == CB_ERRSPACE) {
		delete pComboBoxData;
		return nResult;
	}

	int nTest = CComboBox::SetItemData (nResult, reinterpret_cast<DWORD> (pComboBoxData));
	m_bAllowOrderMod = false;
	return nResult;
}	// AddString


// Action:	Adds a string to a sorted list or one with insertions.
// Throws:	CMemoryException: =, GetLBText
int CSortComboBox::SortStrings (CString& strItem1, CComboBoxData* pComboBoxData1,
								int nCount)
{
	CString strItem2;
	CComboBoxData* pComboBoxData2;

	if (m_bSorted) {
		// Binary search insertion
		int nLower = 0, nUpper = nCount - 1, nMid, nCompare;

		while (nLower <= nUpper) {
			nMid = (nLower + nUpper) / 2;
			GetLBText (nMid, strItem2);
			pComboBoxData2 = GetItemDataObjPtr (nMid);

			nCompare = m_pComboBoxOrder->Compare (strItem1, pComboBoxData1,
				strItem2, pComboBoxData2);

			if (nCompare > 0)
				nLower = nMid + 1;
			else if (nCompare < 0)
				nUpper = nMid - 1;
			else
				break;
		}

		// When a match is not found:
		// nLower <= position <= nUpper + 1 == nLower
		// See http://www.FileJockeySoftware.com/ANO/BinSearchInsert.html.
		return (nCompare == 0 ? nMid : nLower);
	}
	else {
		// Linear search insertion: only considers added strings

		for (int n = 0;  n < nCount;  ++n) {
			pComboBoxData2 = GetItemDataObjPtr (n);

			if (pComboBoxData2 == nullptr)
				return -1;
			else if (pComboBoxData2->WasAdded()) {
				GetLBText (n, strItem2);

				if (m_pComboBoxOrder->Compare (strItem1, pComboBoxData1,
				strItem2, pComboBoxData2) < 0)
					return n;
			}
		}

		return -1;	// after last string
	}
}	// SortStrings


// Action:	Inserts a string into the combo box without sorting.
// Note:	Allows for storing a data item that is in a subclass of CComboBoxData.
int CSortComboBox::InsertString (int nIndex, LPCTSTR lpszItem,
								 CComboBoxData* pComboBoxData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pComboBoxData;
		return CB_ERR;
	}

	if (!CheckComboBoxData (&pComboBoxData))
		return CB_ERR;

	int nResult = CComboBox::InsertString (nIndex, lpszItem);
	if (nResult == CB_ERR || nResult == CB_ERRSPACE) {
		delete pComboBoxData;
		return nResult;
	}

	CComboBox::SetItemData (nResult, reinterpret_cast<DWORD> (pComboBoxData));
	pComboBoxData->SetInserted();
	m_bSorted        = false;
	m_bAllowOrderMod = false;
	return nResult;
}	// InsertString


// Action:	Deletes a string and
//			may allow the sort and string orders to change.
int CSortComboBox::DeleteString(UINT nIndex)
{
	int nResult = CComboBox::DeleteString (nIndex);

	if (GetCount() == 0)
		ResetContent();

	return nResult;
}	// DeleteString


// Action:	Releases memory that is no longer needed.
// Note:	The subclass destructor will be called.
void CSortComboBox::DeleteItem (LPDELETEITEMSTRUCT lpDeleteItemStruct)
{
	delete reinterpret_cast<CComboBoxData*> (lpDeleteItemStruct->itemData);
}	// DeleteItem


// Action:	Releases memory for all items.
// Note:	DeleteItem is not called since item data are set to zero.
void  CSortComboBox::DeleteAllItems()
{
	int nCount = GetCount();
	DWORD dwData;

	for (int n = 0;  n < nCount;  ++n) {
		dwData = CComboBox::GetItemData (n);

		if (dwData != CB_ERR) {
			delete reinterpret_cast<CComboBoxData*> (dwData);
			CComboBox::SetItemData (n, 0);
		}
	}
}	// DeleteAllItems


// Action:	Releases memory, clears the combo box and allows for the
//			sort and string orders to be changed.
void CSortComboBox::ResetContent()
{
	DeleteAllItems();
	CComboBox::ResetContent();
	m_bSorted        = true;
	m_bAllowOrderMod = true;
}	// ResetContent


// Action:	Releases memory when this window is destroyed.
void CSortComboBox::OnDestroy()
{
	DeleteAllItems(); 
	CComboBox::OnDestroy();
}	// OnDestroy


// Action:	Stores user data in a CComboBoxData object.
int CSortComboBox::SetItemData (int nIndex, DWORD dwItemData)
{
	CComboBoxData* pComboBoxData = GetItemDataObjPtr (nIndex);

	if (pComboBoxData != nullptr) {
		pComboBoxData->SetItemData (dwItemData);
		return TRUE;
	}
	else
		return CB_ERR;
}	// SetItemData


// Action:	Retrieves user data from a CComboBoxData object.
DWORD CSortComboBox::GetItemData (int nIndex) const
{
	CComboBoxData* pComboBoxData = GetItemDataObjPtr (nIndex);

	if (pComboBoxData != nullptr)
		return pComboBoxData->GetItemData();
	else
		return CB_ERR;
}	// GetItemData


// Action:	Stores user data in a CComboBoxData object.
int CSortComboBox::SetItemDataPtr (int nIndex, void* pData)
{
	CComboBoxData* pComboBoxData = GetItemDataObjPtr (nIndex);

	if (pComboBoxData != nullptr) {
		pComboBoxData->SetItemData (reinterpret_cast<DWORD>(pData));
		return TRUE;
	}
	else
		return CB_ERR;
}	// SetItemDataPtr


// Action:	Retrieves user data from a CComboBoxData object.
void* CSortComboBox::GetItemDataPtr (int nIndex) const
{
	CComboBoxData* pComboBoxData = GetItemDataObjPtr (nIndex);

	if (pComboBoxData != nullptr)
		return reinterpret_cast<void*> (pComboBoxData->GetItemData());
	else
		return reinterpret_cast<void*> (CB_ERR);
}	// GetItemDataPtr


// Action:	Stores user data as a pointer to a CComboBoxData object.
// Note:	Copies the data values from the old data object and deletes it.
//			pComboBoxData should be from a subclass of CComboBoxData.
int CSortComboBox::SetItemDataObjPtr (int nIndex, CComboBoxData* pComboBoxData)
{
	if (pComboBoxData == nullptr)
		return CB_ERR;

	CComboBoxData* pOldComboBoxData = GetItemDataObjPtr (nIndex);

	if (pComboBoxData != nullptr) {
		pComboBoxData = pOldComboBoxData;
		delete pOldComboBoxData;
		return CComboBox::SetItemData (nIndex, reinterpret_cast<DWORD> (pComboBoxData));
	}
	else
		return CB_ERR;
}	// SetItemDataObjPtr


// Action:	Retrieves user data as a pointer to a CComboBoxData object.
CComboBoxData* CSortComboBox::GetItemDataObjPtr (int nIndex) const
{
	int nCount = GetCount();

	if (nCount == CB_ERR || nIndex < 0 || nIndex >= nCount)
		return nullptr;

	DWORD dwData = CComboBox::GetItemData (nIndex);

	if (dwData != CB_ERR)
		return reinterpret_cast<CComboBoxData*> (dwData);
	else
		return nullptr;
}	// GetItemDataObjPtr
