// SortListCtrl.cpp : implementation file
// Copyright (c) 2016 by David Wincelberg
//
// Window styles:
// Recommended: report view
// Optional: sort ascending or sort descending
//
// Reference:
// Microsoft Systems Journal, December 1996, Wicked Code
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SortListCtrl.h"

#include "ListCtrlData.h"
#include "ListCtrlOrder.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSortListCtrl

// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortListCtrl::CSortListCtrl (CListCtrlOrder* pListCtrlOrder /* = nullptr */,
							  bool bHandleDelete /* = true */)
{
	m_pListCtrlOrder = (pListCtrlOrder == nullptr ? new CListCtrlOrder() : pListCtrlOrder);
	m_bHandleDelete  = bHandleDelete;
	m_bSorting       = false;
	m_nLastCol       = -1;
	m_bReverse       = false;
	m_bResetDir      = true;
	m_bAllowOrderMod = true;
	m_bSuspend       = false;
}	// CSortListCtrl


// Note:	These parameters relate to the Strategy and Flyweight patterns.
CSortListCtrl::CSortListCtrl (CStringOrder* pStringOrder,
							  bool bHandleDeleteInner /* = true */,
							  bool bHandleDeleteOuter /* = true */)
{
	m_pListCtrlOrder = new CListCtrlOrder (pStringOrder, bHandleDeleteInner);
	m_bHandleDelete  = bHandleDeleteOuter;
	m_bSorting       = false;
	m_nLastCol       = -1;
	m_bReverse       = false;
	m_bResetDir      = true;
	m_bAllowOrderMod = true;
}	// CSortListCtrl


CSortListCtrl::~CSortListCtrl()
{
	if (m_bHandleDelete)
		delete m_pListCtrlOrder;
}	// ~CSortListCtrl


BEGIN_MESSAGE_MAP(CSortListCtrl, CListCtrl)
	//{{AFX_MSG_MAP(CSortListCtrl)
	ON_NOTIFY_REFLECT(LVN_GETDISPINFO, OnGetDispInfo)
	ON_NOTIFY_REFLECT(LVN_COLUMNCLICK, OnColumnClick)
	ON_NOTIFY_REFLECT(LVN_DELETEITEM, OnDeleteItem)
	ON_NOTIFY_REFLECT(LVN_DELETEALLITEMS, OnDeleteAllIems)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSortListCtrl message handlers

// Action:	Sets the sorting flags and values.  Removes styles that interfere
//			with custom text display and sorting.
void CSortListCtrl::PreSubclassWindow()
{
	DWORD dwStyle = GetStyle();

	if ((dwStyle & LVS_SORTASCENDING) != 0)
		m_bSorting = true;
	else if ((dwStyle & LVS_SORTDESCENDING) != 0) {
		m_bSorting  = true;
		m_nLastCol  = 0;
		m_bReverse  = true;
		m_bResetDir = false;
	}

	// Removing these styles allows for text callback and a comparison function.
	ModifyStyle (LVS_SORTASCENDING | LVS_SORTDESCENDING, 0);

	CListCtrl::PreSubclassWindow();
}	// PreSubclassWindow


// Action:	Replaces the sort-order object.
//			Deletes the old one unless m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortListCtrl::SetSortOrder (CListCtrlOrder* pListCtrlOrder,
								  bool bHandleDelete /* = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the sort-order strategy is not allowed ")
			_T("after adding an item.\n"));
		ASSERT (false);
		return;
	}

	if (pListCtrlOrder == nullptr)
		return;

	if (m_bHandleDelete)
		delete m_pListCtrlOrder;

	m_pListCtrlOrder = pListCtrlOrder;
	m_bHandleDelete  = bHandleDelete;
}	// SetSortOrder


// Action:	Replaces the string-order object.
//			Deletes the old one unless CListCtrlOrder's m_bHandleDelete is false.
// Note:	This relates to the Strategy and Flyweight patterns.
void CSortListCtrl::SetStringOrder (CStringOrder* pStringOrder,
									bool bHandleDelete /*  = true */)
{
	if (!m_bAllowOrderMod) {
		TRACE (_T("** Changing the string-order strategy is not allowed ")
			_T("after adding an item.\n"));
		ASSERT (false);
		return;
	}

	if (pStringOrder == nullptr || m_pListCtrlOrder == nullptr)
		return;

	m_pListCtrlOrder->SetStringOrder (pStringOrder, bHandleDelete);
}	// SetStringOrder


// Action:	Checks if a listctrl-data object was supplied; o.w. allocates one.
bool CSortListCtrl::CheckListCtrlData (CListCtrlData **ppListCtrlData, LPCTSTR lpszItem,
									   LPARAM lParam /* = 0 */)
{
	if (*ppListCtrlData == nullptr) {
		try {
			*ppListCtrlData = new CListCtrlData (this, lParam);
			(*ppListCtrlData)->SetAt (0, lpszItem);
		}
		catch (CMemoryException* e) {
			e->Delete();
			return false;
		}
	}

	return true;
}	// CheckListCtrlData


// Action:	Inserts an item into a list control.
int CSortListCtrl::InsertItem (int nItem, LPCTSTR lpszItem,
							   CListCtrlData* pListCtrlData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListCtrlData;
		return -1;
	}

	if (!CheckListCtrlData (&pListCtrlData, lpszItem))
		return -1;

	if (nItem == -1)
		nItem = GetItemCount();

	int nResult = CListCtrl::InsertItem (nItem, LPSTR_TEXTCALLBACK);

	if (nResult != -1) {
		CListCtrl::SetItemData (nResult, reinterpret_cast<DWORD> (pListCtrlData));
		m_bAllowOrderMod = false;
		if (m_bSorting)
			Resort();
	}
	else
		delete pListCtrlData;

	return nResult;
}	// InsertItem


// Action:	Inserts an item into a list control.
int CSortListCtrl::InsertItem (int nItem, LPCTSTR lpszItem, int nImage,
							   CListCtrlData* pListCtrlData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListCtrlData;
		return -1;
	}

	if (!CheckListCtrlData (&pListCtrlData, lpszItem))
		return -1;

	if (nItem == -1)
		nItem = GetItemCount();

	int nResult = CListCtrl::InsertItem (nItem, LPSTR_TEXTCALLBACK, nImage);

	if (nResult != -1) {
		CListCtrl::SetItemData (nResult, reinterpret_cast<DWORD> (pListCtrlData));
		m_bAllowOrderMod = false;
		if (m_bSorting)
			Resort();
	}
	else
		delete pListCtrlData;

	return nResult;
}	// InsertItem


// Action:	Inserts an item into a list control.
int CSortListCtrl::InsertItem (UINT nMask, int nItem, LPCTSTR lpszItem, UINT nState,
							   UINT nStateMask, int nImage, LPARAM lParam,
							   CListCtrlData* pListCtrlData /* = nullptr */)
{
	if (lpszItem == nullptr) {
		delete pListCtrlData;
		return -1;
	}

	if (!CheckListCtrlData (&pListCtrlData, lpszItem, lParam))
		return -1;

	if (nItem == -1)
		nItem = GetItemCount();

	// Valid members: pszText and lParam
	nMask |= (LVIF_TEXT | LVIF_PARAM);

	int nResult = CListCtrl::InsertItem (nMask, nItem, LPSTR_TEXTCALLBACK, nState,
		nStateMask, nImage, reinterpret_cast<DWORD> (pListCtrlData));

	if (nResult != -1) {
		m_bAllowOrderMod = false;
		if (m_bSorting)
			Resort();
	}
	else
		delete pListCtrlData;

	return nResult;
}	// InsertItem


// Action:	Inserts an item into a list control.
int CSortListCtrl::InsertItem (LVITEM *pItem,
							   CListCtrlData* pListCtrlData /* = nullptr */)
{
	if (pItem == nullptr || pItem->pszText == nullptr) {
		delete pListCtrlData;
		return -1;
	}

	if (!CheckListCtrlData (&pListCtrlData, pItem->pszText, pItem->lParam))
		return -1;

	// Valid members: pszText and lParam
	pItem->mask   |= (LVIF_TEXT | LVIF_PARAM);
	pItem->pszText = LPSTR_TEXTCALLBACK;
	pItem->lParam  = reinterpret_cast<DWORD> (pListCtrlData);
	pItem->iItem   = (pItem->iItem == -1 ? GetItemCount() : pItem->iItem);

	int nResult = CListCtrl::InsertItem (pItem);

	if (nResult != -1) {
		m_bAllowOrderMod = false;
		if (m_bSorting)
			Resort();
	}
	else
		delete pListCtrlData;

	return nResult;
}	// InsertItem


// Action:	Deletes an item from a list control and
//			may allow the sort and string orders to change.
BOOL CSortListCtrl::DeleteItem (int nItem)
{
	BOOL bResult = CListCtrl::DeleteItem (nItem);

	if (GetItemCount() == 0) {
		LRESULT r;
		OnDeleteAllIems (nullptr, &r);
	}

	return bResult;
}	// DeleteItem


// Action:	Gets the text for items and subitems.
// Note:	lstrcpyn will null-terminate a string that is too long.
void CSortListCtrl::OnGetDispInfo (NMHDR* pNMHDR, LRESULT* pResult)
{
	LV_DISPINFO* pDispInfo = reinterpret_cast<LV_DISPINFO*> (pNMHDR);

	if ((pDispInfo->item.mask & LVIF_TEXT) != 0
	&&   pDispInfo->item.lParam != 0) {
		CListCtrlData* pListCtrlData = reinterpret_cast<CListCtrlData*>
			(pDispInfo->item.lParam);

		::lstrcpyn (pDispInfo->item.pszText,
			pListCtrlData->GetAt (pDispInfo->item.iSubItem), pDispInfo->item.cchTextMax);
	}

	*pResult = 0;
}	// OnGetDispInfo


// Action:	Deletes the stored data when an item is deleted.
// Note:	The subclass destructor will be called.
void CSortListCtrl::OnDeleteItem (NMHDR* pNMHDR, LRESULT* pResult)
{
	NM_LISTVIEW* pNMListView = reinterpret_cast<NM_LISTVIEW*> (pNMHDR);
	delete reinterpret_cast<CListCtrlData*> (pNMListView->lParam);

	*pResult = 0;
}	// OnDeleteItem


// Action:	Allows for changing the sort and string orders.
void CSortListCtrl::OnDeleteAllIems (NMHDR* pNMHDR, LRESULT* pResult)
{
	m_bAllowOrderMod = true;

	if (m_bResetDir) {
		m_nLastCol = -1;
		m_bReverse = false;
	}

	*pResult = 0;
}	// OnDeleteAllIems


// Action:	Supports reverse sorting.
bool CSortListCtrl::GetReverse (int nColumn, bool bUpdate /* = true */)
{
	if (nColumn != m_nLastCol)
		m_bReverse = false;
	else if (bUpdate)
		m_bReverse = !m_bReverse;

	m_nLastCol = nColumn;
	return m_bReverse;
}	// GetReverse


// Action:	Calls CListCtrlOrder::Compare to sort items.
int CALLBACK CSortListCtrl::CompareFunc (LPARAM lParam1, LPARAM lParam2,
										 LPARAM lParamSort)
{
	CListCtrlOrder* pListCtrlOrder = reinterpret_cast<CListCtrlOrder*> (lParamSort);
	return pListCtrlOrder->Compare (lParam1, lParam2);
}	// CompareFunc


// Action:	Sorts the list control using the supplied parameters.
void CSortListCtrl::Sort (int nCol /*  = 0 */, bool bReverse /* = false */)
{
	if (m_pListCtrlOrder == nullptr)
		return;

	m_pListCtrlOrder->SetValues (nCol, bReverse);
	SortItems (CompareFunc, reinterpret_cast<DWORD> (m_pListCtrlOrder));
	SetColSorted (nCol, bReverse);
}	// Sort


// Action:	Signals that a column is sorted so that clicking on a column
//			button would result in the correct response.
void CSortListCtrl::SetColSorted (int nCol /* = 0 */, bool bReverse /* = false */) {
	m_nLastCol = nCol;
	m_bReverse = bReverse;
}	// SetColSorted


// Action:	Resorts the list control using the current sort direction.
void CSortListCtrl::Resort()
{
	if (m_pListCtrlOrder == nullptr)
		return;

	m_pListCtrlOrder->SetValues (0, GetReverse (0, false));
	SortItems (CompareFunc, reinterpret_cast<DWORD> (m_pListCtrlOrder));
}	// Resort


// Action:	Sorts the selected column.  May reverse sort.
void CSortListCtrl::OnColumnClick (NMHDR* pNMHDR, LRESULT* pResult)
{
	if (m_bSuspend || m_pListCtrlOrder == nullptr)
		return;

	NM_LISTVIEW* pNMListView = reinterpret_cast<NM_LISTVIEW*> (pNMHDR);

	m_pListCtrlOrder->SetValues (pNMListView->iSubItem,
		GetReverse (pNMListView->iSubItem));

	SortItems (CompareFunc, reinterpret_cast<DWORD> (m_pListCtrlOrder));

	*pResult = 0;
}	// OnColumnClick


// Action:	Stores user data in a CListCtrlData object.
BOOL CSortListCtrl::SetItemData (int nItem, DWORD dwData)
{
	CListCtrlData* pListCtrlData = GetItemDataObjPtr (nItem);

	if (pListCtrlData != nullptr) {
		pListCtrlData->SetLParam (dwData);
		return TRUE;
	}
	else
		return FALSE;
}	// SetItemData


// Action:	Retrieves user data from a CListCtrlData object.
DWORD CSortListCtrl::GetItemData (int nItem) const
{
	CListCtrlData* pListCtrlData = GetItemDataObjPtr (nItem);

	if (pListCtrlData != nullptr)
		return pListCtrlData->GetLParam();
	else
		return 0;
}	// GetItemData


// Action:	Stores user data as a pointer to a CListCtrlData object.
// Note:	Copies the data values from the old data object and deletes it.
//			pListCtrlData should be from a subclass of CListCtrlData.
BOOL CSortListCtrl::SetItemDataObjPtr (int nItem, CListCtrlData* pListCtrlData)
{
	if (pListCtrlData == nullptr)
		return FALSE;

	CListCtrlData* pOldListCtrlData = GetItemDataObjPtr (nItem);

	if (pOldListCtrlData != nullptr) {
		pListCtrlData = pOldListCtrlData;
		delete pOldListCtrlData;
		return CListCtrl::SetItemData (nItem, reinterpret_cast<DWORD> (pListCtrlData));
	}
	else
		return FALSE;
}	// SetItemDataObjPtr


// Action:	Retrieves user data as a pointer to a subclass of CListCtrlData.
CListCtrlData* CSortListCtrl::GetItemDataObjPtr (int nItem) const
{
	if (nItem < 0 || nItem >= GetItemCount())
		return nullptr;

	DWORD dwData = CListCtrl::GetItemData (nItem);

	return reinterpret_cast<CListCtrlData*> (dwData);
}	// GetItemDataObjPtr
